// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../../dump.h"
#include "../../slice.h"
#include "../../sim.h"
#include "../../ntricks.h"
#include "celiv.h"
#include "../../gui_hooks.h"
struct celiv celiv_config;

static int unused __attribute__((unused));

void sim_celiv(struct device *in)
{
gui_start();
celiv_load_config(&celiv_config,in);



in->dt=celiv_config.celiv_dt;

carrier_count_reset(in);

in->go_time=FALSE;

dump_dynamic_init(in);
//struct istruct celivout;
//inter_init(&celivout);
in->externalv=celiv_config.celiv_Vexternal;
printf("Running celiv %e\n",in->externalv);
int ittr=0;

int step=0;
double dj;
double j0;
light_solve_and_update(in,&(in->mylight),in->Psun,0.0);

in->Vapplied=0.0;
if (fabs(celiv_config.celiv_Vexternal-in->Vapplied)>0.2)
{
	ramp_externalv(in,0.0,celiv_config.celiv_Vexternal);
}

sim_externalv(in,celiv_config.celiv_Vexternal);
in->Vapplied=celiv_config.celiv_Vexternal;


double V=in->Vapplied;
double V0=celiv_config.celiv_Vexternal;
light_solve_and_update(in,&(in->mylight),in->Psun,0.0);

newton_celiv(in,V0,FALSE);
device_timestep(in);
newton_celiv(in,V0,FALSE);
device_timestep(in);

in->time=-celiv_config.celiv_t0;

time_init(in);

if (celiv_config.celiv_enable_time_mesh==TRUE)
{
	time_update_dt(in);
}

device_timestep(in);

in->go_time=TRUE;



//double np_out_tot=0.0;
double t0=0.0;
double t1=celiv_config.celiv_t1;
double t2=celiv_config.celiv_t1+celiv_config.celiv_t2;
double t3=celiv_config.celiv_t1+celiv_config.celiv_t2+celiv_config.celiv_t3;
FILE *celiv_out=fopen("./celiv_j.dat","w");
fclose(celiv_out);

celiv_out=fopen("./celiv_v.dat","w");
fclose(celiv_out);

celiv_out=fopen("./celiv_rtot.dat","w");
fclose(celiv_out);

double ramp_rate=0.0;
double i_max=0.0;
double t_max=0.0;
double mu=0.0;
double A=fabs(celiv_config.celiv_Vmax/(t1-t0));
double np_den=0.0;
struct istruct out_i;
inter_init(&out_i);

struct istruct out_v;
inter_init(&out_v);

struct istruct out_mue;
inter_init(&out_mue);

struct istruct out_muh;
inter_init(&out_muh);

struct istruct out_mue_c;
inter_init(&out_mue_c);

struct istruct out_muh_c;
inter_init(&out_muh_c);

struct istruct out_G;
inter_init(&out_G);

struct istruct lost_charge;
inter_init(&lost_charge);

struct istruct field;
inter_init(&field);

struct istruct celiv_vapplied;
inter_init(&celiv_vapplied);





double mu_e_start=get_avg_mue(in);
double mu_h_start=get_avg_muh(in);

//	printf("%le %le\n",in->Vapplied,in->Vapplied_last);
	//exit(0);
double extracted_through_contacts=0.0;
int i;
//reset_npinit(in);
do
{
ramp_rate=celiv_config.celiv_Vmax*(in->dt/(t1-t0));

	if (in->time<t0)
	{
		V=V0;


		double ratio=light_convert_density(in,celiv_config.celiv_laser_on,celiv_config.celiv_laser_width);

		if (ratio>0.0)
		{
			light_solve_and_update(in,&(in->mylight),in->Psun,ratio*celiv_config.celiv_laser_eff);
		}else
		{
			light_solve_and_update(in,&(in->mylight),in->Psun,0.0);
		}

		for (i=0;i<in->ymeshpoints;i++)
		{
		in->n_orig[i]=in->n[i]+in->nt_all[i];
		in->p_orig[i]=in->p[i]+in->pt_all[i];

		in->n_orig_f[i]=in->n[i];
		in->p_orig_f[i]=in->p[i];

		in->n_orig_t[i]=in->nt_all[i];
		in->p_orig_t[i]=in->pt_all[i];
		}

		extracted_through_contacts=0.0;
	}else
	if (in->time<t1)
	{
		light_solve_and_update(in,&(in->mylight),in->Psun,0.0);
		V+=ramp_rate;

	}else
	{
		V=V0;
	}

	double i0=newton_celiv(in,V,TRUE);

	if ((in->time>t0)&&(in->time<(t1)))
	{
		if (fabs(i0)>i_max)
		{
			i_max=fabs(i0);
			t_max=in->time-t0;
		}
		j0=fabs(in->C*(in->Vapplied-in->Vapplied_last)/in->dt);
		extracted_through_contacts+=-in->dt*get_J(in)/in->ylen/q;
	}

	if (in->time<t0)
	{
		np_den=fabs(get_extracted_n(in)+get_extracted_p(in))/2.0;
	}

	if (get_dump_status(dump_iodump)==TRUE)
	{
		printf("celiv time=%e step=%d %.1e ",in->time,step,in->last_error);
		printf("Vtot=%lf current = %e mA (%e A/m^2)\n",V,get_I(in)/1e-3,get_J(in));
	}
	ittr++;

	plot_now(in,"celiv.plot");

	inter_append(&out_i,in->time,i0);

	if (get_dump_status(dump_iodump)==TRUE)
	{
		inter_append(&out_v,in->time,V);
		inter_append(&out_mue,in->time,get_avg_mue(in));
		inter_append(&out_muh,in->time,get_avg_muh(in));
		inter_append(&out_mue_c,in->time,in->mun[0]);
		inter_append(&out_muh_c,in->time,in->mup[0]);
		inter_append(&out_G,in->time,in->Gn[0]);
		inter_append(&lost_charge,in->time,extracted_through_contacts-fabs(get_extracted_n(in)+get_extracted_p(in))/2.0);
		inter_append(&celiv_vapplied,in->time,in->Vapplied);
		inter_append(&field,in->time,get_avg_field(in));
	}

	dump_dynamic_add_data(in);
	slice_check(in);

	device_timestep(in);
	step++;

	if (in->time>t3) break;

	if (celiv_config.celiv_enable_time_mesh==TRUE)
	{
		time_update_dt(in);
	}

	if (in->stop==TRUE)
	{
		break;

	}
	stop_start(in);

}while(1);

dump_dynamic_save(in);
dump_dynamic_free(in);

char outpath[1000];
if (celiv_config.celiv_invert_current==TRUE) inter_mul(&out_i,-1.0);
sprintf(outpath,"%s%s",in->outputpath,"celiv_i.dat");
inter_save_backup(&out_i,outpath,get_dump_status(dump_iodump));

if (get_dump_status(dump_iodump)==TRUE)
{
	sprintf(outpath,"%s%s",in->outputpath,"celiv_v.dat");
	inter_save(&out_v,outpath);

	sprintf(outpath,"%s%s",in->outputpath,"celiv_mue.dat");
	inter_save(&out_mue,outpath);

	sprintf(outpath,"%s%s",in->outputpath,"celiv_muh.dat");
	inter_save(&out_muh,outpath);

	sprintf(outpath,"%s%s",in->outputpath,"celiv_mue_c.dat");
	inter_save(&out_mue_c,outpath);

	sprintf(outpath,"%s%s",in->outputpath,"celiv_muh_c.dat");
	inter_save(&out_muh_c,outpath);

	sprintf(outpath,"%s%s",in->outputpath,"celiv_G.dat");
	inter_save(&out_G,outpath);

	sprintf(outpath,"%s%s",in->outputpath,"celiv_lost_charge.dat");
	inter_save(&lost_charge,outpath);

	sprintf(outpath,"%s%s",in->outputpath,"celiv_field.dat");
	inter_save(&field,outpath);

	sprintf(outpath,"%s%s",in->outputpath,"celiv_vapplied.dat");
	inter_save(&celiv_vapplied,outpath);

	printf("celiv stats\n");
	//inter_get(&out_i,in->celiv_time_j1)-inter_get(&out_i,in->celiv_time_j0)

	dj=fabs(inter_get(&out_i,celiv_config.celiv_time_j1)-i_max);


	mu=(2.0/3.0)*in->ylen*in->ylen/(A*t_max*t_max*(1.0+0.21*(dj/j0)));
	printf("%le %le\n",in->ylen*in->ylen,(A*t_max*t_max*(1.0+0.21*(dj/j0))));
	printf("t_max=%le i_max=%le mu=%le dj=%le j0=%le A=%le\n",t_max,i_max,mu,dj,j0,A);
	celiv_out=fopena(in->outputpath,"./celiv_mu_deibel.dat","w");
	fprintf(celiv_out,"%le %le %le\n",mu_e_start,mu_h_start,mu);
	fclose(celiv_out);

	mu=(2.0/3.0)*in->ylen*in->ylen/(A*t_max*t_max*(1.0+0.36*(dj/j0)));
	printf("%le %le\n",in->ylen*in->ylen,(A*t_max*t_max*(1.0+0.36*(dj/j0))));
	printf("t_max=%le i_max=%le mu=%le dj=%le j0=%le A=%le\n",t_max,i_max,mu,dj,j0,A);
	celiv_out=fopena(in->outputpath,"./celiv_mu_juska.dat","w");
	fprintf(celiv_out,"%le %le %le\n",mu_e_start,mu_h_start,mu);
	fclose(celiv_out);

	celiv_out=fopena(in->outputpath,"./celiv_ce_eff.dat","w");
	fprintf(celiv_out,"%le %le %le %le %le\n",celiv_config.celiv_Vmax,celiv_config.celiv_laser_eff,(extracted_through_contacts/np_den)*100.0,extracted_through_contacts,np_den);
	fclose(celiv_out);

	celiv_out=fopena(in->outputpath,"./celiv_max_mu.dat","w");
	fprintf(celiv_out,"%le %le\n",t_max+t0,fabs(inter_get(&out_mue,t_max+t0)));
	fclose(celiv_out);
}

in->go_time=FALSE;

inter_free(&out_G);
inter_free(&out_muh);
inter_free(&out_mue);
inter_free(&out_muh_c);
inter_free(&out_mue_c);
inter_free(&out_i);
inter_free(&out_v);
inter_free(&field);
inter_free(&celiv_vapplied);
inter_free(&lost_charge);
gui_stop();
}


void celiv_load_config(struct celiv *in,struct device *dev)
{
double ver=0.0;
char name[200];
FILE *config=fopena(dev->outputpath,"celiv.inp","r");


config_read_line_to_double(&(celiv_config.celiv_dt),config,"#celiv_dt");

config_read_line_to_double(&(in->celiv_t0),config,"#celiv_t0");

config_read_line_to_double(&(in->celiv_t1),config,"#celiv_t1");

config_read_line_to_double(&(in->celiv_t2),config,"#celiv_t2");

config_read_line_to_double(&(in->celiv_Vexternal),config,"#Vexternal");

config_read_line_to_double(&(in->celiv_Vmax),config,"#Vmax");

config_read_line_to_double(&(in->celiv_time_j0),config,"#celiv_time_j0");

config_read_line_to_double(&(in->celiv_time_j1),config,"#celiv_time_j1");

config_read_line_to_double(&(in->Rshort_celiv),config,"#Rshort");

config_read_line_to_double(&(in->celiv_laser_on),config,"#celiv_laser_on");

config_read_line_to_double(&(in->celiv_laser_width),config,"#celiv_laser_width");

config_read_line_to_double(&(in->celiv_laser_eff),config,"#celiv_laser_eff");
in->celiv_laser_eff=fabs(in->celiv_laser_eff);

config_read_line_to_int(&(celiv_config.celiv_enable_time_mesh),config,"#celiv_enable_time_mesh");

config_read_line_to_int(&(in->celiv_start_at_vbi),config,"#solve_at_Vbi");

config_read_line_to_int(&(in->celiv_invert_current),config,"#invert_current");



unused=fscanf(config,"%s",name);
unused=fscanf(config,"%lf",&(ver));
if (ver!=1.22)
{
printf("File compatability problem %s\n","celiv.inp");
exit(0);
}

unused=fscanf(config,"%s",name);
if (strcmp("#end",name)!=0)
{
printf("Problem with celiv.inp file last item read %s!\n",name);
exit(0);
}
fclose(config);

}
