#ifndef device_h
#define device_h
#include "code_ctrl.h"
#include "light.h"

struct mesh
{
double number;
double len;
double den;
};

struct layer
{
double height;
char name[100];
};

struct layers
{
int number;
struct layer* l;
};

struct device
{
	int remesh;
	int newmeshsize;
	double Jnleft;
	double Jnright;
	double Jpleft;
	double Jpright;
	double *phi;
	double *Nad;
	double *G;
	double *Gn;
	double *Gp;
	double *n;
	double *p;
	double *dn;
	double *dndphi;
	double *dp;
	double *dpdphi;
	double *Eg;
	double *Xi;
	double *Ev;
	double *Ec;

	double *mun;
	double *mup;

	double *epsilonr;

	double *Fn;
	double *Fp;
	double *Nc;
	double *Nv;
	double *Tl;
	double *Te;
	double *Th;
	double *ymesh;
	double *R;
	double *Fi;
	int *imat;
	double *Jn;
	double *Jp;
	double Vapplied;
	int ymeshpoints;
	double Vl;
	double Vr;
	double *x;
	double *t;
	double *xp;
	double *tp;
	double *b;
	double *kf;
	double *kd;
	double *kr;

	double *Rbi;
	double *Rn;
	double *Rp;
	double *kl;
	double *ke;
	double *kh;
	double *Hl;
	double *He;
	double *Hh;
	double *Habs;
	double Psun;
	int dd_conv;
	int excite_conv;
	int thermal_conv;
	int newton_enable_external_thermal;

	double deltaFln;
	double deltaFlp;
	double deltaFrn;
	double deltaFrp;

	int mat_layers;

	double *Rbi_k;

	double *ex;
	double *Dex;
	double *Hex;

	double *nfinit;
	double *pfinit;
	double *ntinit;
	double *ptinit;

	double *nfequlib;
	double *pfequlib;
	double *ntequlib;
	double *ptequlib;

	double **ntbinit;
	double **ptbinit;

	double xlen;
	double ylen;
	double zlen;

	double kTq;

	struct layers mat;
	
	int N;
	int M;
	int *Ti;	//row
	int *Tj;	//col
	double *Tx;	//data

	int lr_pcontact;
	int invert_applied_bias;

//plotting
	FILE *gnuplot;
	FILE *gnuplot_time;
	FILE *converge;
	FILE *tconverge;
	int max_electrical_itt;
	double electrical_clamp;
	int max_electrical_itt0;
	double electrical_clamp0;
	double min_cur_error;
	double Pmax_voltage;

//Device characterisation
	double Voc;
	double Jsc;
	double FF;
	double Pmax;
	double Tll;
	double Tlr;
	int Tliso;
	double dt;
	int srh_sim;
	#ifdef enable_time
	int go_time;
	double time;
	double *nlast;
	double *plast;
	#endif
	int ntrapnewton;
	int ptrapnewton;

	int stop;
	double Rshort;
	struct mesh *meshdata;
	int ymeshlayers;
	int onlypos;
	int odes;
	double last_error;
	double posclamp;
	int srh_bands;
	double *wn;
	double *wp;

//n traps
	double *nt_all;
	double **nt;
	#ifdef enable_time
	double **ntlast;
	#endif
	double **dnt;
	double **srh_n_r1;
	double **srh_n_r2;
	double **srh_n_r3;
	double **srh_n_r4;
	double **dsrh_n_r1;
	double **dsrh_n_r2;
	double **dsrh_n_r3;
	double **dsrh_n_r4;
	double **Fnt;
	double **xt;
	double *tt;

	double **nt_r1;
	double **nt_r2;
	double **nt_r3;
	double **nt_r4;
//p traps
	double *pt_all;
	double **pt;
	#ifdef enable_time
	double **ptlast;
	#endif
	double **dpt;
	double **srh_p_r1;
	double **srh_p_r2;
	double **srh_p_r3;
	double **srh_p_r4;
	double **dsrh_p_r1;
	double **dsrh_p_r2;
	double **dsrh_p_r3;
	double **dsrh_p_r4;
	double **Fpt;
	double **xpt;
	double *tpt;

	double **pt_r1;
	double **pt_r2;
	double **pt_r3;
	double **pt_r4;

	double A;
	double Vol;

	double Rshunt;
	double Rcontact;

	int Dphoton;
	int lr_bias;

	int *dostype;

	int interfaceleft;
	int interfaceright;
	double phibleft;
	double phibright;
	double vl_e;
	double vl_h;
	double vr_e;
	double vr_h;
	int stop_start;
	double externalv;
	double Vapplied_last;
	double Ilast;
	int plottime;
	int timedumpcount;
	char outputpath[200];
	int simmode;
	double area;

	double *nrelax;
	double *ntrap_to_p;
	double *prelax;
	double *ptrap_to_n;



	double lcharge;
	double rcharge;



	int dumpitdos;


	double t_big_offset;

	double C;
	double other_layers;
	int last_ittr;

	int newton_cur;
	void (*newton_aux)(struct device* ,double ,double* ,double* ,double* ,double* ,double* ,double* ,double* ,double*);
	double B;
	double xnl_left;
	double xpl_left;
	int stoppoint;
	double ilast;

	int newton_clever_exit;
	int slave;
	char slave_path[100];
	char plot_file[100];
	int master;

	double start_stop_time;


	double Is;
	double n_id;
	double Igen;
	struct light mylight;
	int dump_movie;

	double *n_orig;
	double *p_orig;
	double *n_orig_f;
	double *p_orig_f;
	double *n_orig_t;
	double *p_orig_t;
	int nofluxl;

	double Vbi;
	int enabled;
	int newton_min_itt;
	double vbi;
	double avg_gen;
	int dump_slicepos;
};
#endif
