// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <string.h>
#include "../../plugins/fit/fit.h"
#include "../../sim.h"
double lib_error(struct fitvars *config,int ext,char * path)
{
//printf("get_tpc_error\n");
char exp_name[40];
char sim_name[40];
char config_name[40];
char out_name[40];
char temp[100];
sprintf(exp_name,"%s./exp/tpc%d/tpc.dat",path,ext);
sprintf(sim_name,"%s./sim/tpc%d/tpcfit.dat",path,ext);
sprintf(out_name,"%s./error_tpc%d.dat",path,ext);
sprintf(config_name,"%s./exp/tpc%d/tpcfit.inp",path,ext);

double error=0.0;
double max=0.0;
struct istruct exp;
struct istruct simtime;
double start=-6.7;
double stop=-4;
double error_mul=1.0;
inter_load_a(&exp,exp_name,0,1);
inter_load_a(&simtime,sim_name,0,1);

inter_chop(&exp,1e-7, 1.0);
inter_chop(&simtime,1e-7, 1.0);

FILE *file_config;
if ((file_config=fopen(config_name,"r"))==NULL)
{
    printf("Error opening file: %s\n", config_name);
    getchar();
}

fscanf(file_config,"%s",temp);
fscanf(file_config,"%le",&(start));
fscanf(file_config,"%s",temp);
fscanf(file_config,"%le",&(stop));
fscanf(file_config,"%s",temp);
fscanf(file_config,"%le",&(error_mul));
fclose(file_config);

double min=inter_get_min(&simtime);

inter_sub_double(&simtime,min);
inter_sub_double(&exp,min);

inter_purge_zero(&simtime);
inter_purge_zero(&exp);

inter_log_y(&simtime);
inter_log_y(&exp);
inter_log_x(&simtime);
inter_log_x(&exp);

inter_chop(&exp,start, stop);
inter_chop(&simtime,start, stop);

min=exp.data[exp.len-1];

inter_sub_double(&simtime,min);
inter_sub_double(&exp,min);

start=exp.x[0];
if (start<simtime.x[0]) start=simtime.x[0];
stop=exp.x[exp.len-1];


max=inter_norm_to_one_range(&exp,start,stop);
inter_rescale(&simtime,1.0, 1.0/max);

//inter_save(&exp,"./a.dat");
//inter_save(&simtime,"./b.dat");

error=inter_get_diff(NULL,&exp,&simtime,start,stop,error_mul);

inter_free(&exp);
inter_free(&simtime);
//exit(0);
//printf("get_tpc_error ret= %d %e %e %e\n",ext,error,min,max );
return error;
}
