# General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
# base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
# The model can simulate OLEDs, Perovskite cells, and OFETs.
# 
# Copyright (C) 2008-2020 Roderick C. I. MacKenzie
# All rights reserved.
# 
# https://www.gpvdm.com
# r.c.i.mackenzie at googlemail.com
# 
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the GPVDM nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import pygtk
pygtk.require('2.0')
import gtk
import sys
import os
import shutil
from numpy import *
from matplotlib.figure import Figure
from numpy import arange, sin, pi
from matplotlib.backends.backend_gtkagg import FigureCanvasGTKAgg as FigureCanvas
import gobject
from util import find_data_file

(
  COLUMN_LAYER,
  COLUMN_THICKNES,
  COLUMN_EDITABLE
) = range(3)

(
  MESH_THICKNES,
  MESH_POINTS,
  MESH_EDITABLE
) = range(3)


articles = []
mesh_articles = []

class tab_electrical_mesh(gtk.Window):
	lines=[]
	edit_list=[]

	line_number=[]
	save_file_name=""

	file_name=""
	name=""
	visible=1

	def __create_model(self):

		# create list store
		model = gtk.ListStore(
		    gobject.TYPE_STRING,
		    gobject.TYPE_STRING,
		    gobject.TYPE_BOOLEAN
		)

		# add items
		for item in articles:
			iter = model.append()

			model.set (iter,
			  COLUMN_LAYER, item[COLUMN_LAYER],
			  COLUMN_THICKNES, item[COLUMN_THICKNES],
			  COLUMN_EDITABLE, item[COLUMN_EDITABLE]
			)

		return model

	def __create_model_mesh(self):

		model_mesh = gtk.ListStore(
		    gobject.TYPE_STRING,
		    gobject.TYPE_STRING,
		    gobject.TYPE_BOOLEAN
		)


		for item in mesh_articles:
			iter = model_mesh.append()

			model_mesh.set (iter,
			  MESH_THICKNES, item[MESH_THICKNES],
			  MESH_POINTS, item[MESH_POINTS],
			  MESH_EDITABLE, item[MESH_EDITABLE]
			)
		return model_mesh

	def __add_columns(self, treeview):

		model = treeview.get_model()

		# Layer tag
		renderer = gtk.CellRendererText()
		renderer.connect("edited", self.on_cell_edited, model)
		renderer.set_data("column", COLUMN_LAYER)

		column = gtk.TreeViewColumn("Layer", renderer, text=COLUMN_LAYER,
				       editable=COLUMN_EDITABLE)
		treeview.append_column(column)

		# Thicknes
		renderer = gtk.CellRendererText()
		renderer.connect("edited", self.on_cell_edited, model)
		renderer.set_data("column", COLUMN_THICKNES)

		column = gtk.TreeViewColumn("Thicknes", renderer, text=COLUMN_THICKNES,
				       editable=COLUMN_EDITABLE)
		treeview.append_column(column)


	def __add_columns_mesh(self, treeview):

		model = treeview.get_model()

		# Thicknes
		renderer = gtk.CellRendererText()
		renderer.connect("edited", self.on_mesh_edited, model)
		renderer.set_data("column", MESH_THICKNES)

		column = gtk.TreeViewColumn("Thicknes", renderer, text=MESH_THICKNES,
				       editable=MESH_EDITABLE)
		treeview.append_column(column)

		# Thicknes
		renderer = gtk.CellRendererText()
		renderer.connect("edited", self.on_mesh_edited, model)
		renderer.set_data("column", MESH_POINTS)

		column = gtk.TreeViewColumn("Points", renderer, text=MESH_POINTS,
				       editable=MESH_EDITABLE)
		treeview.append_column(column)



	def on_add_item_clicked(self, button, model):
		new_item = ["#layerN","0e-9" ,True]
		articles.append(new_item)

		iter = model.append()
		model.set (iter,
		    COLUMN_LAYER, new_item[COLUMN_LAYER],
		    COLUMN_THICKNES, new_item[COLUMN_THICKNES],
		    COLUMN_EDITABLE, new_item[COLUMN_EDITABLE]
		)

	def on_add_mesh_clicked(self, button, model):
		new_item = ["0e-9","0" ,True]
		mesh_articles.append(new_item)

		iter = model.append()
		model.set (iter,
		    MESH_THICKNES, new_item[MESH_THICKNES],
		    MESH_POINTS, new_item[MESH_POINTS],
		    MESH_EDITABLE, new_item[MESH_EDITABLE]
		)

	def save_model(self, ):
		print "Saved"
		a = open("device_epitaxy.inp", "w")
		a.write("#layers\n")
		a.write(str(len(self.layer_model))+"\n")


		for item in self.layer_model:
			a.write(item[COLUMN_LAYER]+"\n")
			a.write(item[COLUMN_THICKNES]+"\n")
		a.write("#mesh\n")
		a.write(str(len(self.mesh_model))+"\n")
		for item in self.mesh_model:
			a.write(item[MESH_THICKNES]+"\n")
			a.write(item[MESH_POINTS]+"\n")
		a.write("#ver\n")
		a.write("1.0\n")
		a.write("#end\n")
		a.close()

	def on_remove_item_clicked(self, button, treeview):

		selection = treeview.get_selection()
		model, iter = selection.get_selected()

		if iter:
			path = model.get_path(iter)[0]
			model.remove(iter)

			del articles[ path ]

		self.save_model()

	def on_remove_from_mesh_click(self, button, treeview):

		selection = treeview.get_selection()
		model, iter = selection.get_selected()

		if iter:
			path = model.get_path(iter)[0]
			model.remove(iter)

			del mesh_articles[ path ]

		self.save_model()

	def on_cell_edited(self, cell, path_string, new_text, model):

		iter = model.get_iter_from_string(path_string)
		path = model.get_path(iter)[0]
		column = cell.get_data("column")

		if column == COLUMN_LAYER:
			articles[path][COLUMN_LAYER] = new_text

			model.set(iter, column, articles[path][COLUMN_LAYER])

		if column == COLUMN_THICKNES:
			#old_text = model.get_value(iter, column)
			articles[path][COLUMN_THICKNES] = new_text
			print new_text
			model.set(iter, column, articles[path][COLUMN_THICKNES])

		self.save_model()

	def on_mesh_edited(self, cell, path_string, new_text, model):

		iter = model.get_iter_from_string(path_string)
		path = model.get_path(iter)[0]
		column = cell.get_data("column")
		print "new text",new_text
		print "path=",path


		if column == MESH_THICKNES:
			mesh_articles[path][MESH_THICKNES] = new_text

			model.set(iter, column, mesh_articles[path][MESH_THICKNES])

		if column == MESH_POINTS:
			#old_text = model.get_value(iter, column)
			mesh_articles[path][MESH_POINTS] = new_text
			print new_text
			model.set(iter, column, mesh_articles[path][MESH_POINTS])

		self.save_model()
		self.update_graph()

	def update_graph(self):
		cmd = self.exe_command+' --onlypos'
		ret= os.system(cmd)
		self.fig.clf()
		self.draw_graph()
		self.fig.canvas.draw()

	def draw_graph(self):

		print "Drawing graph"

		n=0

		ax1 = self.fig.add_subplot(111)
		#ax2 = ax1.twinx()
		x_pos=0.0
		layer=0
		color =['r','g','b','y','o','r','g','b','y','o']

		ax1.set_ylabel('Energy (eV)')
		#ax2.set_ylabel('Energy (eV)')
		ax1.set_xlabel('Position (nm)')
		try:
			t,s = loadtxt("equ_Ec.dat", unpack=True)
			t=t*1e9
			Ec, = ax1.plot(t,s, 'ro-', linewidth=3 ,alpha=0.5)

			t,s = loadtxt("equ_Ev.dat", unpack=True)
			t=t*1e9
			Ev,=ax1.plot(t,s, 'go-', linewidth=3 ,alpha=0.5)

			t,s = loadtxt("equ_Fi.dat", unpack=True)
			t=t*1e9
			Fi,=ax1.plot(t,s, 'bo-', linewidth=3 ,alpha=0.5)

			self.fig.legend((Ec, Ev, Fi), ('LUMO', 'HOMO', 'Fi'), 'upper right')

		except:
			print "No mode file\n"
	def save_image(self,file_name):
		self.fig.savefig(file_name)

	def callback_refresh(self, widget, data=None):
		self.update_graph()


	def callback_close(self, widget, data=None):
		self.hide()
		return True

	def callback_save(self, widget, data=None):
		dialog = gtk.FileChooserDialog("Save as..",
                               None,
                               gtk.FILE_CHOOSER_ACTION_SAVE,
                               (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                gtk.STOCK_SAVE, gtk.RESPONSE_OK))
		dialog.set_default_response(gtk.RESPONSE_OK)

		filter = gtk.FileFilter()
		filter.set_name(".jpg")
		filter.add_pattern("*.jpg")
		dialog.add_filter(filter)

		response = dialog.run()
		if response == gtk.RESPONSE_OK:
			file_name=dialog.get_filename()

			print os.path.splitext(file_name)[1]
			if os.path.splitext(file_name)[1]:
				self.save_image(file_name)
			else:
				filter=dialog.get_filter()
				self.save_image(file_name+filter.get_name())

		elif response == gtk.RESPONSE_CANCEL:
		    print 'Closed, no files selected'
		dialog.destroy()

	def callback_help(self, widget, data=None):
		cmd = 'firefox http://www.roderickmackenzie.eu/opvdm_wiki.html'
		os.system(cmd)

	def wow(self,exe_command):
		self.hbox=gtk.HBox()
		self.exe_command=exe_command
		self.edit_list=[]
		self.line_number=[]
		self.save_file_name="device_epitaxy.inp"
		print "loading ",self.save_file_name
		f = open(self.save_file_name)
		self.lines = f.readlines()
		f.close()
		pos=0
		for i in range(0, len(self.lines)):
			self.lines[i]=self.lines[i].rstrip()

		n=0
		gui_pos=0
		pos=0
		pos=pos+1	#first comment
		layers=int(self.lines[pos])

		for i in range(0, layers):
			pos=pos+1
			label=self.lines[pos]	#read label
			print label

			pos=pos+1
			layer_ticknes=self.lines[pos] 	#read thicknes

			articles.append([ label, str(layer_ticknes), True ])

		pos=pos+1
		pos=pos+1
		mesh_layers=int(self.lines[pos])

		for i in range(0, mesh_layers):
			pos=pos+1
			thicknes=self.lines[pos]	#read label
			print label

			pos=pos+1
			points=self.lines[pos] 	#read thicknes

			mesh_articles.append([ str(thicknes), str(points), True ])



			n=n+1
		gui_pos=gui_pos+1

		tooltips = gtk.Tooltips()

		toolbar = gtk.Toolbar()
		#toolbar.set_orientation(gtk.ORIENTATION_VERTICAL)
		toolbar.set_style(gtk.TOOLBAR_ICONS)
		toolbar.set_size_request(-1, 50)

		tool_bar_pos=0
		save = gtk.ToolButton(gtk.STOCK_SAVE)
		tooltips.set_tip(save, "Save image")
		save.connect("clicked", self.callback_save)
		toolbar.insert(save, tool_bar_pos)
		tool_bar_pos=tool_bar_pos+1

		save = gtk.ToolButton(gtk.STOCK_REFRESH)
		tooltips.set_tip(save, "Refresh mesh")
		save.connect("clicked", self.callback_refresh)
		toolbar.insert(save, tool_bar_pos)
		tool_bar_pos=tool_bar_pos+1

		sep = gtk.SeparatorToolItem()
		sep.set_draw(False)
		sep.set_expand(True)
		toolbar.insert(sep, tool_bar_pos)
		sep.show()
		tool_bar_pos=tool_bar_pos+1

		help = gtk.ToolButton(gtk.STOCK_HELP)
		toolbar.insert(help, tool_bar_pos)
		help.connect("clicked", self.callback_help)
		help.show()
		tool_bar_pos=tool_bar_pos+1

		close = gtk.ToolButton(gtk.STOCK_QUIT)
		close.connect("clicked", self.callback_close)
		toolbar.insert(close, tool_bar_pos)
		close.show()
		tool_bar_pos=tool_bar_pos+1

		toolbar.show_all()
		window_main_vbox=gtk.VBox()
		window_main_vbox.pack_start(toolbar, False, True, 0)
		#self.attach(toolbar, 0, 1, 0, 1)
		tool_bar_pos=tool_bar_pos+1



		self.fig = Figure(figsize=(5,4), dpi=100)
		self.fig.subplots_adjust(bottom=0.2)
		self.fig.subplots_adjust(left=0.1)

		self.draw_graph()
		canvas = FigureCanvas(self.fig)  # a gtk.DrawingArea
		#canvas.set_background('white')
		#canvas.set_facecolor('white')
		canvas.figure.patch.set_facecolor('white')
		canvas.set_size_request(500, 150)
		canvas.show()
		self.hbox.pack_start(canvas, True, True, 0)
		#self.attach(canvas, 1, 3, 0, 1)


		#Layer editor
		self.layer_model = self.__create_model()
		self.mesh_model = self.__create_model_mesh()

	        vbox = gtk.VBox(False, 2)


		frame = gtk.Frame()
		frame.set_label("Layers")
		vbox_layers = gtk.VBox(False, 2)
		treeview = gtk.TreeView(self.layer_model)
		treeview.set_size_request(200, 100)
		treeview.set_rules_hint(True)
		treeview.get_selection().set_mode(gtk.SELECTION_SINGLE)
		self.__add_columns(treeview)
	        vbox_layers.pack_start(treeview, False, False, 0)
		treeview.show()



		if os.path.exists("./server.inp")==True:
			add_button = gtk.Button("Add",gtk.STOCK_ADD)
			add_button.connect("clicked", self.on_add_item_clicked, self.layer_model)
			add_button.show()

			delete_button = gtk.Button("Delete",gtk.STOCK_DELETE)
			delete_button.connect("clicked", self.on_remove_item_clicked, treeview)
			delete_button.show()


			hbox = gtk.HBox(False, 2)
			hbox.pack_start(add_button, False, False, 0)
			hbox.pack_start(delete_button, False, False, 0)
			hbox.show()
			vbox_layers.pack_start(hbox, False, False, 0)

		vbox_layers.show()

		frame.add(vbox_layers)
		frame.show()
	        vbox.pack_start(frame, False, False, 0)

		#spacer
		label=gtk.Label(" \n\n    ")
		#self.attach(label, 4, 5, 1, 2,gtk.SHRINK ,gtk.SHRINK)
		vbox.pack_start(label, False, False, 0)

		label.show()



		#mesh editor
		frame = gtk.Frame()
		frame.set_label("Mesh")
		vbox_mesh = gtk.VBox(False, 2)
		treeview = gtk.TreeView(self.mesh_model)
		treeview.set_size_request(200, 100)
		treeview.set_rules_hint(True)
		treeview.get_selection().set_mode(gtk.SELECTION_SINGLE)
		self.__add_columns_mesh(treeview)
		vbox_mesh.pack_start(treeview, False, False, 0)
		treeview.show()

		gui_pos=3
		add_button = gtk.Button("Add",gtk.STOCK_ADD)
		add_button.connect("clicked", self.on_add_mesh_clicked, self.mesh_model)
		add_button.show()

		delete_button = gtk.Button("Delete",gtk.STOCK_DELETE)
		delete_button.connect("clicked", self.on_remove_from_mesh_click, treeview)
		delete_button.show()

	        hbox = gtk.HBox(False, 2)

	        hbox.pack_start(add_button, False, False, 0)
	        hbox.pack_start(delete_button, False, False, 0)
		vbox_mesh.pack_start(hbox, False, False, 0)
		frame.add(vbox_mesh)
		vbox.pack_start(frame, False, False, 0)
		frame.show()
		vbox_mesh.show()
		hbox.show()
		self.hbox.pack_start(vbox, False, False, 0)
		#self.attach(vbox, 3, 4, 0, 1,gtk.SHRINK ,gtk.SHRINK)
		vbox.show()
		window_main_vbox.add(self.hbox)
		self.add(window_main_vbox)
		self.set_title("Electrical Mesh Editor - (www.opvdm.com)")
		self.set_icon_from_file(find_data_file("gui/mesh.png"))
		self.connect("delete-event", self.callback_close)
		self.set_position(gtk.WIN_POS_CENTER)
		self.show_all()


