# General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
# base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
# The model can simulate OLEDs, Perovskite cells, and OFETs.
# 
# Copyright (C) 2008-2020 Roderick C. I. MacKenzie
# All rights reserved.
# 
# https://www.gpvdm.com
# r.c.i.mackenzie at googlemail.com
# 
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the GPVDM nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import sys
import pygtk
inp_dir='/usr/share/opvdm/'
gui_dir='/usr/share/opvdm/gui/'
lib_dir='/usr/lib64/opvdm/'
sys.path.append('./gui/')
sys.path.append(lib_dir)

import os
import shutil
import signal
import subprocess
from util import replace
from util import get_token_value

import os, fnmatch
import stat

def copy_check_ver(orig,file_name,dest,only_over_write,clever):
	#remove the dest file if both exist ready to copy
	do_copy=True
	orig_ver=""
	dest_ver=""
	lines_src=[]
	lines_dest=[]
	#if we are only over writing only copy if dest file exists
	if (only_over_write==True):
		if os.path.exists(dest+file_name)==False:
			print file_name," - only exists in the source archive"
			do_copy=False
			return

	if clever==True:
		#if the src file exists get it's ver number
		if os.path.exists(orig+file_name)==True:
			f = open(orig+file_name)
			lines_src = f.readlines()
			f.close()

			for i in range(0, len(lines_src)):
				lines_src[i]=lines_src[i].rstrip()

			if lines_src.count("#ver")>0:
				i=lines_src.index('#ver')+1
				orig_ver=lines_src[i]

		#if the dest file exists get it's ver number
		if os.path.exists(dest+file_name)==True:
			f = open(dest+file_name)
			lines_dest = f.readlines()
			f.close()

			for i in range(0, len(lines_dest)):
				lines_dest[i]=lines_dest[i].rstrip()

			if lines_dest.count('#ver')>0:
				i=lines_dest.index('#ver')+1
				dest_ver=lines_dest[i]

		#if if they are not equal
		if orig_ver!=dest_ver:
			print "Verstion numbers do not match for files",orig+file_name,orig_ver,dest+file_name,dest_ver
			for i in range(0,len(lines_dest)):
				if lines_dest[i].startswith("#") and lines_dest[i]!="#ver" and lines_dest[i]!="#end":
					lookfor=lines_dest[i]
					for ii in range(0,len(lines_src)):
						if lines_src[ii]==lookfor:
							print "Found",lines_dest[i],lines_src[ii]
							lines_dest[i+1]=lines_src[ii+1]
							break

			a = open(dest+file_name, "w")
			for i in range(0,len(lines_dest)):
				a.write(lines_dest[i]+"\n")
			a.close()

			return

	if (do_copy==True):
		if os.path.exists(dest+file_name):
			if os.path.exists(orig+file_name):
				os.remove(dest+file_name)

		if os.path.exists(orig+file_name):
			shutil.copyfile(orig+file_name, dest+file_name)
			print file_name," - copied"

def import_archive(file_name,dest_dir,only_over_write):
	if file_name.endswith('.tar.gz')==True:
		tmp_dir="/tmp/opvdm"

		if os.path.exists(tmp_dir):
			shutil.rmtree(tmp_dir)

		if not os.path.exists(tmp_dir):
			os.makedirs(tmp_dir)

		cmd = 'tar -xzf '+file_name+' -C '+tmp_dir+'/'
		os.system(cmd)

		pattern='sim.inp'
		path='/tmp/opvdm/'
		res=""
		for root, dirs, files in os.walk(path):
			for name in files:
				if fnmatch.fnmatch(name, pattern):
					res=os.path.join(root, name)
					break
			if res!="":
				break

		sim_path=os.path.dirname(res)+"/"
	else:
		sim_path=file_name+"/"


	files=[ "sim.inp", "device.inp", "stark.inp" ,"shg.inp" ,"dos0.inp" ,"jv.inp" ,"celiv.inp" , "optics.inp", "math.inp",  "dump.inp" , "light.inp", "tpv.inp", "otrace.inp", "server.inp", "pulse_voc.inp","pulse.inp","light_exp.inp" ]

	for i in files:
		copy_check_ver(sim_path,i,dest_dir,only_over_write,True)


	files=[ "device_epitaxy.inp", "optics_epitaxy.inp", "fit.inp", "dump_time.inp", "constraints.inp","duplicate.inp","dump_time.inp", "thermal.inp" ]

	for i in files:
		copy_check_ver(sim_path,i,dest_dir,only_over_write,False)


