// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../../dump.h"
#include "../../slice.h"
#include "../../sim.h"
#include "../../ntricks.h"
#include "otrace.h"
struct otrace otrace_config;

static int unused __attribute__((unused));

void sim_otrace(struct device *in)
{
otrace_load_config(&otrace_config,in);
struct istruct tpv_trans;



sim_tpv(in);
in->go_time=FALSE;
inter_load_a(&tpv_trans,"./tpv_voc.dat",0,1);
otrace_config.otrace_Vexternal=tpv_trans.data[0];
//<clean>



in->dt=otrace_config.otrace_dt;

carrier_count_reset(in);

in->go_time=FALSE;

dump_dynamic_init(in);
//struct istruct otraceout;
//inter_init(&otraceout);
in->externalv=otrace_config.otrace_Vexternal;
printf("Running otrace %e\n",in->externalv);
int ittr=0;

int step=0;
double dj;
double j0;
light_solve_and_update(in,&(in->mylight),in->Psun,0.0);


sim_externalv(in,otrace_config.otrace_Vexternal);
in->Vapplied=otrace_config.otrace_Vexternal;


double V=in->Vapplied;
double V0=otrace_config.otrace_Vexternal;
light_solve_and_update(in,&(in->mylight),in->Psun,0.0);

newton_otrace(in,V0,FALSE);
device_timestep(in);
newton_otrace(in,V0,FALSE);
device_timestep(in);

in->time=-otrace_config.otrace_t0;

time_init(in);

if (otrace_config.otrace_enable_time_mesh==TRUE)
{
	time_update_dt(in);
}

device_timestep(in);

in->go_time=TRUE;



//double np_out_tot=0.0;
double t0=0.0;
double t1=otrace_config.otrace_t1;
double t2=otrace_config.otrace_t1+otrace_config.otrace_t2;
double t3=otrace_config.otrace_t1+otrace_config.otrace_t2+otrace_config.otrace_t3;
FILE *otrace_out=fopen("./otrace_j.dat","w");
fclose(otrace_out);

otrace_out=fopen("./otrace_v.dat","w");
fclose(otrace_out);

otrace_out=fopen("./otrace_rtot.dat","w");
fclose(otrace_out);

double ramp_rate=0.0;
double i_max=0.0;
double t_max=0.0;
double mu=0.0;
double A=fabs(otrace_config.otrace_Vmax/(t1-t0));
double np_den=0.0;
struct istruct out_i;
inter_init(&out_i);

struct istruct out_v;
inter_init(&out_v);

struct istruct out_mue;
inter_init(&out_mue);

struct istruct out_muh;
inter_init(&out_muh);

struct istruct out_mue_c;
inter_init(&out_mue_c);

struct istruct out_muh_c;
inter_init(&out_muh_c);

struct istruct out_G;
inter_init(&out_G);

struct istruct lost_charge;
inter_init(&lost_charge);

struct istruct field;
inter_init(&field);

struct istruct otrace_vapplied;
inter_init(&otrace_vapplied);



double mu_e_start=get_avg_mue(in);
double mu_h_start=get_avg_muh(in);

//	printf("%le %le\n",in->Vapplied,in->Vapplied_last);
	//exit(0);
double extracted_through_contacts=0.0;
int i;
//reset_npinit(in);
do
{

ramp_rate=otrace_config.otrace_Vmax*(in->dt/(t2-t1));

	if (in->time<t0)
	{
		V=V0;


		double ratio=light_convert_density(in,otrace_config.otrace_laser_on,otrace_config.otrace_laser_width);
		//printf("ratio=%le\n",ratio*otrace_config.otrace_laser_eff);
		if (ratio>0.0)
		{
			light_solve_and_update(in,&(in->mylight),in->Psun,ratio*otrace_config.otrace_laser_eff);
		}else
		{
			light_solve_and_update(in,&(in->mylight),in->Psun,0.0);
		}

		for (i=0;i<in->ymeshpoints;i++)
		{
		in->n_orig[i]=in->n[i]+in->nt_all[i];
		in->p_orig[i]=in->p[i]+in->pt_all[i];

		in->n_orig_f[i]=in->n[i];
		in->p_orig_f[i]=in->p[i];

		in->n_orig_t[i]=in->nt_all[i];
		in->p_orig_t[i]=in->pt_all[i];
		}

		extracted_through_contacts=0.0;
	}else
	if (in->time<t1)
	{

			light_solve_and_update(in,&(in->mylight),0.0,0.0);
			if (otrace_config.otrace_match_tpv==TRUE) V=inter_get(&tpv_trans,in->time-t0);
			printf("tpv time = %le\n",in->time-t0);


	}else
	if (in->time<t2)
	{

			V+=ramp_rate;
	}else
	{
		V=V0;
	}

	double i0=-newton_otrace(in,V,TRUE);

	if ((in->time>t0)&&(in->time<(t1)))
	{
		if (i0>i_max)
		{
			i_max=i0;
			t_max=in->time-t0;
		}
		j0=fabs(in->C*(in->Vapplied-in->Vapplied_last)/in->dt);
		extracted_through_contacts+=-in->dt*get_J(in)/in->ylen/q;
	}

	if (in->time<t0)
	{
		np_den=fabs(get_extracted_n(in)+get_extracted_p(in))/2.0;
	}

	if (get_dump_status(dump_iodump)==TRUE)
	{
		printf("otrace time=%e step=%d %.1e ",in->time,step,in->last_error);
		printf("Vtot=%lf current = %e mA (%e A/m^2)\n",V,get_I(in)/1e-3,get_J(in));
	}

	ittr++;

	plot_now(in,"otrace_bands.plot");

	inter_append(&out_i,in->time,i0);
	inter_append(&out_v,in->time,V);
	inter_append(&out_mue,in->time,get_avg_mue(in));
	inter_append(&out_muh,in->time,get_avg_muh(in));
	inter_append(&out_mue_c,in->time,in->mun[0]);
	inter_append(&out_muh_c,in->time,in->mup[0]);
	inter_append(&out_G,in->time,in->Gn[0]);
	inter_append(&lost_charge,in->time,extracted_through_contacts-fabs(get_extracted_n(in)+get_extracted_p(in))/2.0);
	inter_append(&otrace_vapplied,in->time,in->Vapplied);
	inter_append(&field,in->time,get_avg_field(in));
	dump_dynamic_add_data(in);
	slice_check(in);

	device_timestep(in);
	step++;

	if (in->time>t3) break;

	if (otrace_config.otrace_enable_time_mesh==TRUE)
	{
		time_update_dt(in);
	}

	if (in->stop==TRUE)
	{
		break;

	}


}while(1);

dump_dynamic_save(in);
dump_dynamic_free(in);

char outpath[1000];
sprintf(outpath,"%s%s",in->outputpath,"otrace_i.dat");
inter_save(&out_i,outpath);

sprintf(outpath,"%s%s",in->outputpath,"otrace_v.dat");
inter_save(&out_v,outpath);

sprintf(outpath,"%s%s",in->outputpath,"otrace_mue.dat");
inter_save(&out_mue,outpath);

sprintf(outpath,"%s%s",in->outputpath,"otrace_muh.dat");
inter_save(&out_muh,outpath);

sprintf(outpath,"%s%s",in->outputpath,"otrace_mue_c.dat");
inter_save(&out_mue_c,outpath);

sprintf(outpath,"%s%s",in->outputpath,"otrace_muh_c.dat");
inter_save(&out_muh_c,outpath);

sprintf(outpath,"%s%s",in->outputpath,"otrace_G.dat");
inter_save(&out_G,outpath);

sprintf(outpath,"%s%s",in->outputpath,"otrace_lost_charge.dat");
inter_save(&lost_charge,outpath);

sprintf(outpath,"%s%s",in->outputpath,"otrace_field.dat");
inter_save(&field,outpath);

sprintf(outpath,"%s%s",in->outputpath,"otrace_vapplied.dat");
inter_save(&otrace_vapplied,outpath);

printf("otrace stats\n");
//inter_get(&out_i,in->otrace_time_j1)-inter_get(&out_i,in->otrace_time_j0)

dj=fabs(inter_get(&out_i,otrace_config.otrace_time_j1)-i_max);


mu=(2.0/3.0)*in->ylen*in->ylen/(A*t_max*t_max*(1.0+0.21*(dj/j0)));
printf("%le %le\n",in->ylen*in->ylen,(A*t_max*t_max*(1.0+0.21*(dj/j0))));
printf("t_max=%le i_max=%le mu=%le dj=%le j0=%le A=%le\n",t_max,i_max,mu,dj,j0,A);
otrace_out=fopena(in->outputpath,"./otrace_mu_deibel.dat","w");
fprintf(otrace_out,"%le %le %le\n",mu_e_start,mu_h_start,mu);
fclose(otrace_out);

mu=(2.0/3.0)*in->ylen*in->ylen/(A*t_max*t_max*(1.0+0.36*(dj/j0)));
printf("%le %le\n",in->ylen*in->ylen,(A*t_max*t_max*(1.0+0.36*(dj/j0))));
printf("t_max=%le i_max=%le mu=%le dj=%le j0=%le A=%le\n",t_max,i_max,mu,dj,j0,A);
otrace_out=fopena(in->outputpath,"./otrace_mu_juska.dat","w");
fprintf(otrace_out,"%le %le %le\n",mu_e_start,mu_h_start,mu);
fclose(otrace_out);

otrace_out=fopena(in->outputpath,"./otrace_ce_eff.dat","w");
fprintf(otrace_out,"%le %le %le %le %le\n",otrace_config.otrace_Vmax,otrace_config.otrace_laser_eff,(extracted_through_contacts/np_den)*100.0,extracted_through_contacts,np_den);
fclose(otrace_out);

otrace_out=fopena(in->outputpath,"./otrace_max_mu.dat","w");
fprintf(otrace_out,"%le %le\n",t_max+t0,fabs(inter_get(&out_mue,t_max+t0)));
fclose(otrace_out);


in->go_time=FALSE;

inter_free(&out_G);
inter_free(&out_muh);
inter_free(&out_mue);
inter_free(&out_muh_c);
inter_free(&out_mue_c);
inter_free(&out_i);
inter_free(&out_v);
inter_free(&field);
inter_free(&otrace_vapplied);

inter_free(&tpv_trans);

}


void otrace_load_config(struct otrace *in,struct device *dev)
{
double ver=0.0;
char name[200];
FILE *config=fopena(dev->outputpath,"otrace.inp","r");


config_read_line_to_double(&(otrace_config.otrace_dt),config,"#otrace_dt");

config_read_line_to_double(&(in->otrace_t0),config,"#otrace_t0");

config_read_line_to_double(&(in->otrace_t1),config,"#otrace_t1");

config_read_line_to_double(&(in->otrace_t2),config,"#otrace_t2");

config_read_line_to_double(&(in->otrace_t3),config,"#otrace_t3");

config_read_line_to_double(&(in->otrace_Vexternal),config,"#Vexternal");

config_read_line_to_double(&(in->otrace_Vmax),config,"#Vmax");

config_read_line_to_double(&(in->otrace_time_j0),config,"#otrace_time_j0");

config_read_line_to_double(&(in->otrace_time_j1),config,"#otrace_time_j1");

config_read_line_to_double(&(in->Rshort_otrace),config,"#Rshort");

config_read_line_to_double(&(in->otrace_laser_on),config,"#otrace_laser_on");

config_read_line_to_double(&(in->otrace_laser_width),config,"#otrace_laser_width");

config_read_line_to_double(&(in->otrace_laser_eff),config,"#otrace_laser_eff");
in->otrace_laser_eff=fabs(in->otrace_laser_eff);

config_read_line_to_int(&(otrace_config.otrace_enable_time_mesh),config,"#otrace_enable_time_mesh");

unused=fscanf(config,"%s",name);
unused=fscanf(config,"%d",&(in->otrace_start_at_vbi));

config_read_line_to_int(&(in->otrace_match_tpv),config,"#otrace_match_tpv");


unused=fscanf(config,"%s",name);
unused=fscanf(config,"%lf",&(ver));
if (ver!=1.21)
{
ewe("File compatability problem %s\n","otrace.inp");
}

unused=fscanf(config,"%s",name);
if (strcmp("#end",name)!=0)
{
ewe("Problem with otrace.inp file last item read %s!\n",name);
}
fclose(config);

}
