// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../../solver.h"
#include "../../sim.h"
double min_error=1e-6;
#define xpoints 1000
static double x[xpoints];
static double V[xpoints];
static double Vp[xpoints];
static double Vpp[xpoints];
static double dt=0.0;
static double dx=0.0;



void wire_addtime()
{
int i;
for (i=0;i<xpoints;i++)
{
	Vpp[i]=Vp[i];
	Vp[i]=V[i];
}
}

void dump(double *x,double *V)
{
int i;
FILE *out;
out=fopen("./wave.dat","w");
for (i=0;i<xpoints;i++)
{
	fprintf(out,"%e %e\n",x[i],V[i]);
}
fclose(out);

}

double get_error(double *V)
{

double error=0.0;

int i;

for (i=0;i<xpoints;i++)
{
	error+=mod(V[i]);
}

return error;
}

double wire_get_voltage_at_end()
{
return Vp[xpoints-1];
}

void wire_init(double in,double timestep)
{

FILE *out=fopen("./wire.dat","w");
fclose(out);
int i=0;
double xpos=0.0;
double len=20.0;
dx=len/((double)xpoints);
for (i=0;i<xpoints;i++)
{
	V[i]=in;
	Vp[i]=in;
	Vpp[i]=in;
	x[i]=xpos;
	xpos+=dx;
}
dt=timestep;
}

void wire_set_dt(double dt_step)
{
dt=dt_step;
}

double wire_sim(double Vleft)
{
//printf("Solve wave\n");
int i;
int N=xpoints*3-2;

int M=xpoints;
int *Ti =    malloc(N*sizeof(int));
int *Tj =    malloc(N*sizeof(int));
double *Tx = malloc(N*sizeof(double));
double *b = malloc(M*sizeof(double));



double L=3.281*0.059e-6;
double C=3.281*24.6e-12;

double lc=L*C;

double Vl=0.0;
double Vc=0.0;
double Vr=0.0;



double dVl=0.0;
double dVc=0.0;
double dVr=0.0;


double dV=0.0;
double xl=0.0;
double xc=0.0;
double xr=0.0;
double dxl=0.0;
double dxc=0.0;
double dxr=0.0;
double dtl=0.0;
double dtc=0.0;
double dtr=0.0;

int dotime=TRUE;
double R=2.0*(2/1000)*3.281;
double G=0.0;//14.0/1000.0;
int ittr=0;
double error=0.0;
int pos=0;

	do
	{

		pos=0;
		for (i=0;i<xpoints;i++)
		{

			if (i==0)
			{
				Vl=Vleft;

				xl=x[0]-dx;
			}else
			{
				Vl=V[i-1];
				xl=x[i-1];
			}

			if (i==(xpoints-1))
			{
				Vr=V[i];
				xr=x[xpoints-1]+dx;
			}else
			{
				Vr=V[i+1];
				xr=x[i+1];
			}

			Vc=V[i];
			xc=x[i];

			dxl=xc-xl;
			dxr=xr-xc;
			dxc=dxl/2.0+dxr/2.0;

			dVl=1.0/dxl/dxc;
			dVc=-(1.0/dxl/dxc+1.0/dxr/dxc);
			dVr=1.0/dxr/dxc;

			dV=dVl*Vl+dVc*Vc+dVr*Vr;

			if (i==(xpoints-1)) dVc+=dVr;

			if (dotime==TRUE)
			{
				dtl=0.0;
				dtc=0.0;
				dtr=0.0;

				dtl=lc/dt/dt;
				dtc=-(lc/dt/dt+lc/dt/dt);
				dtr=lc/dt/dt;

				dVc+=-dtr;
				dVc+=-R*C/dt-R*G-L*G/dt;


				dV-=dtl*Vpp[i]+dtc*Vp[i]+dtr*V[i]+R*C*(V[i]-Vp[i])/dt+R*G*V[i]+L*G*(V[i]-Vp[i])/dt;
			}

			if (i!=0)
			{
				Ti[pos]=i;
				Tj[pos]=i-1;
				Tx[pos]=dVl;
				pos++;


			}

			//lattice
			Ti[pos]=i;
			Tj[pos]=i;
			Tx[pos]=dVc;
			pos++;


			if (i!=(xpoints-1))
			{
				Ti[pos]=i;
				Tj[pos]=i+1;
				Tx[pos]=dVr;
				pos++;

			}

			b[i]=-(dV);

		}

		error=get_error(b);
		solver(M,N,Ti,Tj, Tx,b);


		for (i=0;i<xpoints;i++)
		{
			V[i]+=b[i];
		}



		printf("Wave error = %e\n",error);
		ittr++;

	}while((ittr<1000)&&(error>min_error));
		//getchar();


	FILE *out;
	out=fopen("./wave.dat","w");
	for (i=0;i<xpoints-1;i++)
	{
		fprintf(out,"%e %e\n",x[i],V[i]);
	}
	fclose(out);


	double cur=0.0;
	for (i=0;i<xpoints-1;i++)
	{
		cur+=dxc*(C*(V[i]-Vp[i])/dt);
	}



free(Ti);
free(Tj);
free(Tx);
free(b);
return cur;
}

/*int wire()
{
wire_init(0.1,1e-9);

FILE *curout=fopen("./curout.dat","w");
int step=0;
double cur=0.0;
double Vleft=0.1;
int count=0;
do
{
Vleft=0.1;
if (count<100) Vleft=0;
cur=wire_sim(Vleft);
printf("%e %e %e %e\n",time,Vp[xpoints-1],cur,Vleft/cur);
wire_addtime();
if (count>200) count=0;
count++;
//getchar();
}while(step<1000);

fclose(curout);

return 0;

}*/

double wire_getvoltage()
{
return V[xpoints-1];
}

